(** Types and modules *)

open Common

type t = {
    expo      : float array array;
    coef      : float array array;
    norm_coef : float array array;
    norm_coef_scale : float array;
    contr     : Contracted_shell.t array;
    index     : int;
    center    : Coordinate.t;
    ang_mom   : Angular_momentum.t;
  }

module Am = Angular_momentum
module Co = Coordinate
module Cs = Contracted_shell



(** Access *)

let ang_mom           t = t.ang_mom
let center            t = t.center
let coefficients      t = t.coef
let contracted_shells t = t.contr
let exponents         t = t.expo
let index             t = t.index
let normalizations    t = t.norm_coef
let norm_scales       t = t.norm_coef_scale
let size_of_shell     t = Array.length t.norm_coef_scale
let size              t = Array.length t.contr


let make ?(index=0) contr =
  assert (Array.length contr > 0);

  let coef = Array.map Cs.coefficients contr
  and expo = Array.map Cs.exponents contr
  in

  let center = Cs.center contr.(0) in
  let rec unique_center = function
    | 0 -> true
    | i -> if Cs.center contr.(i) = center then (unique_center [@tailcall]) (i-1) else false
  in
  if not (unique_center (Array.length contr - 1)) then
    invalid_arg "ContractedAtomicShell.make Coordinate.t differ";

  let ang_mom = Cs.ang_mom contr.(0) in
  let rec unique_angmom = function
    | 0 -> true
    | i -> if Cs.ang_mom contr.(i) = ang_mom then (unique_angmom [@tailcall]) (i-1) else false
  in
  if not (unique_angmom (Array.length contr - 1)) then
    invalid_arg "Contracted_shell.make: Angular_momentum.t differ";

  let norm_coef =
    Array.map Cs.normalizations contr
  in
  let norm_coef_scale = Cs.norm_scales contr.(0)
  in
  { index ; expo ; coef ; center ; ang_mom ; norm_coef ;
    norm_coef_scale ; contr  }


let with_index a i =
  { a with index = i }


(** Printers *)

let pp ppf s =
  let open Format in
  fprintf ppf "@[%3d-%-3d@]" (s.index+1) (s.index+ (size_of_shell s)*(size s));
  fprintf ppf "@[%a@ %a@] @[" Am.pp_string s.ang_mom Co.pp s.center;
  Array.iter2 (fun e_arr c_arr -> fprintf ppf "@[<v>";
                                  Array.iter2 (fun e c -> fprintf ppf "@[%16.8e  %16.8e@]@;" e c)
                                    e_arr c_arr;
                                  fprintf ppf "@;@]") s.expo s.coef;
  fprintf ppf "@]"
